local M = {}

local delivery_route_planner = require('/lua/ge/extensions/gameplay/route/route')()
local helper = require('scenario/scenariohelper')

local max_delivery_distance = 2500
local pizza_shop_pos = vec3(1053.89319, 1323.19946, 145.01397)

local deliver_pizza_msgs = {"The pizza is getting cold Mario!", "What are you waiting for Mario? Get moving!", "We don't have all day Mario!"}
local running_late_msgs = {"You're running late Mario! Get back here!", "Where are you Mario! Get back here!"}
local thanks_msgs = {"Grazie!", "What took you so long!", "Keep the change!", "Why is my pizza covered in soda!", "Where's the pineapple?!", "Thanks for nothing!"}

local buildings_for_delivery = {}

local state = "scenario_started"

local the_building_pos = vec3()
local destination_pos = vec3()

local the_wp1_pos = vec3()
local the_wp2_pos = vec3()

local deliveries_made = 0

local current_delivery_distance = 0
local next_delivery_distance = 0

local time_left = 1
local countdown_display_timer = 1

-- Edited function from groundMarkers.lua since it doesn't work...
local function getPathLength(path)
  local dist = 0
  local pathLength = tableSize(path)
  for i = 1, (pathLength - 1) do
    dist = dist + (path[i + 1].pos - path[i].pos):length()
  end
  return dist
end

local function getRandomBuilding()
	return buildings_for_delivery[math.random(#buildings_for_delivery)]
end

local function getClosestRoadPos(building_pos)
	local wp1, wp2, lat_dist_from_wp = map.findClosestRoad(building_pos)
	
	local wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	local wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local xnorm = clamp(building_pos:xnormOnLine(wp1_pos, wp2_pos), 0, 1)
  local wp_dir = wp2_pos - wp1_pos
  local closest_pos_on_road_to_building = xnorm * wp_dir + wp1_pos
	
	local road_width = map.getMap().nodes[wp1].radius
	
	local offset_vec = (road_width - 1.5) * (building_pos - closest_pos_on_road_to_building):normalized()
	
	return closest_pos_on_road_to_building + offset_vec, wp1, wp2, wp_dir
end

local function getDeliveryTime(dist)
	local avg_speed = (1.4 + deliveries_made * 0.1) * math.pow(dist, 1 / 3) 
	
	return dist / avg_speed
end

local function setDestination(pos, trigger)
	local wp1, wp2, wp_dir = nil
	
	destination_pos, wp1, wp2, wp_dir = getClosestRoadPos(pos)
	
	the_wp1_pos = vec3(map.getMap().nodes[wp1].pos)
	the_wp2_pos = vec3(map.getMap().nodes[wp2].pos)
	
	local rot = quatFromDir(wp_dir, vec3(0,0,1))
	
	if trigger then
		trigger:setPosRot(destination_pos.x, destination_pos.y, destination_pos.z, rot.x, rot.y, rot.z, rot.w)
	end
	
	core_groundMarkers.setFocus(wp1)
	--core_groundMarkers.setFocus(wp1, math.huge) -- Sets navigation to waypoint and removes ground arrow markers
	--core_groundMarkers.onSerialize() -- Gets rid of floating directional arrows
end

local function setupNextDelivery()
	local rand_building = getRandomBuilding()

	the_building_pos = vec3(rand_building:getPosition())	
	delivery_route_planner:setupPathMulti({pizza_shop_pos, the_building_pos})
	next_delivery_distance = getPathLength(delivery_route_planner.path)
end

local function goDeliverPizza(running_late)
	local delivery_point_trigger = scenetree.findObject("delivery_point_trigger")
	setDestination(the_building_pos, delivery_point_trigger)

	current_delivery_distance = next_delivery_distance

	--print("Delivery distance = " .. current_delivery_distance)

	if not running_late then
		local delivery_time = getDeliveryTime(current_delivery_distance)
		
		time_left = time_left + delivery_time	
	end
	
	local rand_msg = deliver_pizza_msgs[math.random(#deliver_pizza_msgs)]
	
	helper.flashUiMessage(rand_msg .. " You have " .. tostring(math.ceil(time_left)) .. " seconds to get there!", 5)
	guihooks.message("Current Objective: Deliver the pizza", 10)

	state = "deliver_pizza"
end

local function goBackToPizzaShop()
	setDestination(pizza_shop_pos)
	
	--print("Back to shop distance = " .. current_delivery_distance)
	local delivery_time = getDeliveryTime(current_delivery_distance)
	time_left = time_left + delivery_time
	
	local rand_msg = thanks_msgs[math.random(#thanks_msgs)]
	
	helper.flashUiMessage(rand_msg, 3)
	guihooks.message("Current Objective: Go back to pizza shop", 10)
	
	deliveries_made = deliveries_made + 1
	
	state = "retrieve_pizza"
end

local function runningLateGoingToPizzaShop()
	local delivery_time = getDeliveryTime(next_delivery_distance)
	time_left = time_left + delivery_time

	local rand_msg = running_late_msgs[math.random(#running_late_msgs)]
	
	helper.flashUiMessage(rand_msg, 5)
	
	state = "retrieve_pizza_late"
end

local function onPreRender(dt, dtSim)
	-- Cylinder to indicate destination
	debugDrawer:drawCylinder((destination_pos + vec3(0, 0, -5)):toPoint3F(), (destination_pos + vec3(0, 0, 75)):toPoint3F(), 1, ColorF(1,0,0,0.25))
	
	local veh = be:getPlayerVehicle(0)
	if not veh or state == "scenario_started" or not be:getEnabled() then return end
	
	if state == "deliver_pizza" or state == "retrieve_pizza_late" then	
		if time_left <= 0 then
			-- Game over
			scenario_scenarios.finish({failed = "The pizza became inedible!"})
			
		elseif time_left <= 10 then
			-- When 10 seconds are left: every second display countdown timer
			
			if countdown_display_timer >= 1 then
				helper.flashUiMessage(tostring(math.ceil(time_left)), 0.5, true)	
				
				countdown_display_timer = 0
			end
			countdown_display_timer = countdown_display_timer + dt
			
		else
			countdown_display_timer = 1
			
		end
	
	elseif state == "retrieve_pizza" then		
		if time_left <= 10 then
			-- If 10 seconds are left before expected arrival time to pizza shop
			-- display message that you are running late!
		
			runningLateGoingToPizzaShop()
		end
	end
	
	time_left = time_left - dt
	
	--print(time_left)
	
	--debugDrawer:drawCylinder((the_building_pos):toPoint3F(), (the_building_pos + vec3(0, 0, 20)):toPoint3F(), 1, ColorF(0,1,0,1))
	--debugDrawer:drawCylinder((the_wp1_pos):toPoint3F(), (the_wp1_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
	--debugDrawer:drawCylinder((the_wp2_pos):toPoint3F(), (the_wp2_pos + vec3(0, 0, 5)):toPoint3F(), 1, ColorF(1,0,0,1))
end

local function onBeamNGTrigger(data)
	if data.subjectName ~= "thePlayer" then return end -- Only care about player on triggers
	
	local stopped = vec3(be:getPlayerVehicle(0):getVelocity()):length() < 0.5
	
	if stopped then
		if data.triggerName == "delivery_point_trigger" then
			if state == "deliver_pizza" then
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goBackToPizzaShop()
				setupNextDelivery()
			end
		elseif data.triggerName == "pizza_shop_trigger" then
			local running_late = state == "retrieve_pizza_late" 	
				
			if state == "retrieve_pizza" or running_late then		
				Engine.Audio.playOnce('AudioGui', "event:UI_Checkpoint")
				goDeliverPizza(running_late)
			end
		end
	end
end

local function getDistancesToBuildings(buildings)
	local arr = {}
	
	local player_pos = vec3(be:getPlayerVehicle(0):getPosition())
	
	for key, object in pairs(buildings) do
		local building_pos = vec3(object:getPosition())
		
		local distance = (player_pos - building_pos):length()
		
		table.insert(arr, {object = object, distance = distance})
	end
	
	--sortDistances(arr)
	
	return arr
end

local function onScenarioLoaded(scenario)
	--local buildings = Sim.findObject('BUILDINGS'):getObjects()
	local objects = getObjectsByClass('TSStatic')

	local datas = getDistancesToBuildings(objects)
	
	for key, data in pairs(datas) do
		local object = data.object
		local distance = data.distance
		
		-- Filtering out bus stops/signs and bridges
		if distance < max_delivery_distance and object.shapeName:match("levels/italy/art/shapes/buildings") then	
			if object.shapeName:match("house") 
			or object.shapeName:match("town_bld") 
			then
				table.insert(buildings_for_delivery, object)
			end
		end
	end
	
	--[[
	for key, object in pairs(objects) do
		-- Filtering out bus stops/signs and bridges
		if object.shapeName:match("levels/italy/art/shapes/buildings") then	
			if object.shapeName:match("house") 
			or object.shapeName:match("town_bld") 
			then
				table.insert(buildings_for_delivery, object)
			end
		end
	end
	]]--
	
	gameplay_traffic.setupTraffic(getMaxVehicleAmount())
	gameplay_traffic.setTrafficVars({spawnValue = 1.3, spawnDirBias = 0.3})
end
	
local function onRaceStart()
	setupNextDelivery()
	goDeliverPizza(false)
end

local function onScenarioRestarted(scenario)
  core_groundMarkers.resetAll()
	countdown_display_timer = 1
	time_left = 1
	deliveries_made = 0
	
	state = "scenario_started"
end

M.onPreRender = onPreRender
M.onBeamNGTrigger = onBeamNGTrigger
M.onScenarioLoaded = onScenarioLoaded
M.onRaceStart = onRaceStart
M.onScenarioRestarted = onScenarioRestarted

return M